@@ -18,31 +18,32 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
                                 datatype, int nprocs,
                                 int myrank, ADIOI_Access
                                 * others_req, ADIO_Offset * offset_list,
-                                ADIO_Offset * len_list, int contig_access_count,
+                                ADIO_Offset * len_list, MPI_Count contig_access_count,
                                 ADIO_Offset
                                 min_st_offset, ADIO_Offset fd_size,
                                 ADIO_Offset * fd_start, ADIO_Offset * fd_end,
-                                MPI_Aint * buf_idx, int *error_code);
+                                MPI_Aint * buf_idx, MPI_Status * status, int *error_code);
 static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                                   * flat_buf, ADIO_Offset * offset_list, ADIO_Offset
-                                  * len_list, int *send_size, int *recv_size,
-                                  int *count, int *start_pos,
-                                  int *partial_send,
-                                  int *recd_from_proc, int nprocs,
+                                  * len_list, MPI_Count * send_size, MPI_Count * recv_size,
+                                  MPI_Count * count, MPI_Count * start_pos,
+                                  MPI_Count * partial_send,
+                                  MPI_Count * recd_from_proc, int nprocs,
                                   int myrank, int
-                                  buftype_is_contig, int contig_access_count,
+                                  buftype_is_contig, MPI_Count contig_access_count,
                                   ADIO_Offset min_st_offset,
                                   ADIO_Offset fd_size,
                                   ADIO_Offset * fd_start, ADIO_Offset * fd_end,
                                   ADIOI_Access * others_req,
-                                  int iter, MPI_Aint buftype_extent, MPI_Aint * buf_idx);
+                                  int iter, MPI_Aint buftype_extent, MPI_Aint * buf_idx,
+                                  MPI_Aint * actual_recved_bytes);
 void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                             * flat_buf, char **recv_buf, ADIO_Offset
                             * offset_list, ADIO_Offset * len_list,
-                            unsigned *recv_size,
+                            MPI_Count * recv_size,
                             MPI_Request * requests, MPI_Status * statuses,
-                            int *recd_from_proc, int nprocs,
-                            int contig_access_count,
+                            MPI_Count * recd_from_proc, int nprocs,
+                            MPI_Count contig_access_count,
                             ADIO_Offset min_st_offset,
                             ADIO_Offset fd_size, ADIO_Offset * fd_start,
                             ADIO_Offset * fd_end, MPI_Aint buftype_extent);
@@ -68,7 +69,7 @@ if ((fd->file_system == ADIO_LUSTRE) && (fd->hints->fs_hints.lustre.lock_ahead_r
 #define ADIOI_LUSTRE_RD_LOCK_AHEAD(fd,cb_nodes,offset,error_code)
 #endif
 
-void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
+void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, MPI_Aint count,
                                MPI_Datatype datatype, int file_ptr_type,
                                ADIO_Offset offset, ADIO_Status * status, int
                                *error_code)
@@ -87,20 +88,17 @@ void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
     /* array of nprocs structures, one for each other process
      * whose request lies in this process's file domain. */
 
-    int i, filetype_is_contig, nprocs, nprocs_for_coll, myrank;
-    int contig_access_count = 0, interleave_count = 0, buftype_is_contig;
-    int *count_my_req_per_proc, count_my_req_procs;
-    int *count_others_req_per_proc, count_others_req_procs;
+    int filetype_is_contig, nprocs, nprocs_for_coll, myrank;
+    MPI_Count contig_access_count = 0;
+    int interleave_count = 0, buftype_is_contig;
+    MPI_Count *count_my_req_per_proc, count_my_req_procs;
+    MPI_Count *count_others_req_per_proc, count_others_req_procs;
     ADIO_Offset start_offset, end_offset, orig_fp, fd_size, min_st_offset, off;
     ADIO_Offset *offset_list = NULL, *st_offsets = NULL, *fd_start = NULL,
         *fd_end = NULL, *end_offsets = NULL;
     ADIO_Offset *len_list = NULL;
     MPI_Aint *buf_idx = NULL;
 
-#ifdef HAVE_STATUS_SET_BYTES
-    MPI_Count bufsize, size;
-#endif
-
     if (fd->hints->cb_pfr != ADIOI_HINT_DISABLE) {
         ADIOI_IOStridedColl(fd, buf, count, ADIOI_READ, datatype,
                             file_ptr_type, offset, status, error_code);
@@ -128,7 +126,7 @@ void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
                               &end_offset, &contig_access_count);
 
 #ifdef RDCOLL_DEBUG
-        for (i = 0; i < contig_access_count; i++) {
+        for (MPI_Count i = 0; i < contig_access_count; i++) {
             DBG_FPRINTF(stderr, "rank %d  off %lld  len %lld\n",
                         myrank, (long long) offset_list[i], (long long) len_list[i]);
         }
@@ -145,7 +143,7 @@ void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
         MPI_Allgather(&end_offset, 1, ADIO_OFFSET, end_offsets, 1, ADIO_OFFSET, fd->comm);
 
         /* are the accesses of different processes interleaved? */
-        for (i = 1; i < nprocs; i++)
+        for (int i = 1; i < nprocs; i++)
             if ((st_offsets[i] < end_offsets[i - 1]) && (st_offsets[i] <= end_offsets[i]))
                 interleave_count++;
         /* This is a rudimentary check for interleaving, but should suffice
@@ -233,7 +231,7 @@ void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
     ADIOI_Read_and_exch(fd, buf, datatype, nprocs, myrank,
                         others_req, offset_list,
                         len_list, contig_access_count, min_st_offset,
-                        fd_size, fd_start, fd_end, buf_idx, error_code);
+                        fd_size, fd_start, fd_end, buf_idx, status, error_code);
 
 
     /* free all memory allocated for collective I/O */
@@ -244,28 +242,19 @@ void ADIOI_GEN_ReadStridedColl(ADIO_File fd, void *buf, int count,
     ADIOI_Free(st_offsets);
     ADIOI_Free(fd_start);
 
-#ifdef HAVE_STATUS_SET_BYTES
-    MPI_Type_size_x(datatype, &size);
-    bufsize = size * count;
-    MPIR_Status_set_bytes(status, datatype, bufsize);
-/* This is a temporary way of filling in status. The right way is to
-   keep track of how much data was actually read and placed in buf
-   during collective I/O. */
-#endif
-
     fd->fp_sys_posn = -1;       /* set it to null. */
 }
 
-void ADIOI_Calc_my_off_len(ADIO_File fd, int bufcount, MPI_Datatype
+void ADIOI_Calc_my_off_len(ADIO_File fd, MPI_Aint bufcount, MPI_Datatype
                            datatype, int file_ptr_type, ADIO_Offset
                            offset, ADIO_Offset ** offset_list_ptr, ADIO_Offset
                            ** len_list_ptr, ADIO_Offset * start_offset_ptr,
-                           ADIO_Offset * end_offset_ptr, int
-                           *contig_access_count_ptr)
+                           ADIO_Offset * end_offset_ptr, MPI_Count * contig_access_count_ptr)
 {
     MPI_Count filetype_size, etype_size;
     MPI_Count buftype_size;
-    int i, j, k;
+    int i, j;
+    MPI_Count k;
     ADIO_Offset i_offset;
     ADIO_Offset frd_size = 0, old_frd_size = 0;
     int st_index = 0;
@@ -273,7 +262,8 @@ void ADIOI_Calc_my_off_len(ADIO_File fd, int bufcount, MPI_Datatype
     ADIO_Offset abs_off_in_filetype = 0;
     ADIO_Offset bufsize;
     ADIO_Offset sum, n_etypes_in_filetype, size_in_filetype;
-    int contig_access_count, filetype_is_contig;
+    int filetype_is_contig;
+    MPI_Count contig_access_count;
     ADIO_Offset *len_list;
     MPI_Aint filetype_extent, filetype_lb;
     ADIOI_Flatlist_node *flat_file;
@@ -477,10 +467,10 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
                                 datatype, int nprocs,
                                 int myrank, ADIOI_Access
                                 * others_req, ADIO_Offset * offset_list,
-                                ADIO_Offset * len_list, int contig_access_count, ADIO_Offset
-                                min_st_offset, ADIO_Offset fd_size,
+                                ADIO_Offset * len_list, MPI_Count contig_access_count,
+                                ADIO_Offset min_st_offset, ADIO_Offset fd_size,
                                 ADIO_Offset * fd_start, ADIO_Offset * fd_end,
-                                MPI_Aint * buf_idx, int *error_code)
+                                MPI_Aint * buf_idx, MPI_Status * status, int *error_code)
 {
 /* Read in sizes of no more than coll_bufsize, an info parameter.
    Send data to appropriate processes.
@@ -492,18 +482,18 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
    array from a file, where each local array is 8Mbytes, requiring
    at least another 8Mbytes of temp space is unacceptable. */
 
-    int i, j, m, ntimes, max_ntimes, buftype_is_contig;
-    ADIO_Offset st_loc = -1, end_loc = -1, off, done, real_off, req_off;
+    int i, m, ntimes, max_ntimes, buftype_is_contig;
+    ADIO_Offset st_loc = -1, end_loc = -1, off, done, real_off;
     char *read_buf = NULL, *tmp_buf;
-    int *curr_offlen_ptr, *count, *send_size, *recv_size;
-    int *partial_send, *recd_from_proc, *start_pos;
+    MPI_Count *curr_offlen_ptr, *count, *send_size, *recv_size;
+    MPI_Count *partial_send, *recd_from_proc, *start_pos;
     /* Not convinced end_loc-st_loc couldn't be > int, so make these offsets */
     ADIO_Offset real_size, size, for_curr_iter, for_next_iter;
-    int req_len, flag, rank;
-    MPI_Status status;
+    int rank;
     ADIOI_Flatlist_node *flat_buf = NULL;
     MPI_Aint lb, buftype_extent;
-    int coll_bufsize;
+    MPI_Aint coll_bufsize;
+    MPI_Aint actual_recved_bytes = 0;
 
     *error_code = MPI_SUCCESS;  /* changed below if error */
     /* only I/O errors are currently reported */
@@ -526,7 +516,7 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
 
     /* now find the real values */
     for (i = 0; i < nprocs; i++)
-        for (j = 0; j < others_req[i].count; j++) {
+        for (MPI_Count j = 0; j < others_req[i].count; j++) {
             st_loc = MPL_MIN(st_loc, others_req[i].offsets[j]);
             end_loc = MPL_MAX(end_loc, (others_req[i].offsets[j]
                                         + others_req[i].lens[j] - 1));
@@ -549,7 +539,7 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
 
     read_buf = fd->io_buf;      /* Allocated at open time */
 
-    curr_offlen_ptr = (int *) ADIOI_Calloc(nprocs * 7, sizeof(int));
+    curr_offlen_ptr = ADIOI_Calloc(nprocs * 7, sizeof(*curr_offlen_ptr));
     /* its use is explained below. calloc initializes to 0. */
 
     count = curr_offlen_ptr + nprocs;
@@ -626,7 +616,41 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
          * minus what was satisfied in previous iteration
          * req_size = size corresponding to req_off */
 
-        size = MPL_MIN((unsigned) coll_bufsize, end_loc - st_loc + 1 - done);
+        size = MPL_MIN(coll_bufsize, end_loc - st_loc + 1 - done);
+        bool flag = false;
+        for (i = 0; i < nprocs; i++) {
+            if (others_req[i].count) {
+                for (MPI_Count j = curr_offlen_ptr[i]; j < others_req[i].count; j++) {
+                    ADIO_Offset req_off;
+                    if (partial_send[i]) {
+                        req_off = others_req[i].offsets[j] + partial_send[i];
+                    } else {
+                        req_off = others_req[i].offsets[j];
+                    }
+                    if (req_off < off + size) {
+                        flag = true;
+                    }
+                }
+            }
+        }
+        if (flag) {
+            MPI_Status read_status;
+            ADIOI_Assert(size == (int) size);
+            ADIOI_LUSTRE_RD_LOCK_AHEAD(fd, fd->hints->cb_nodes, off, error_code);
+            ADIO_ReadContig(fd, read_buf + for_curr_iter, (int) size, MPI_BYTE,
+                            ADIO_EXPLICIT_OFFSET, off, &read_status, error_code);
+            if (*error_code != MPI_SUCCESS) {
+                /* TODO: proper error return */
+                return;
+            }
+            int actual_size;
+            MPI_Get_count(&read_status, MPI_BYTE, &actual_size);
+            if (actual_size < size) {
+                size = actual_size;
+                /* TODO: need abort the further rounds */
+            }
+        }
+
         real_off = off - for_curr_iter;
         real_size = size + for_curr_iter;
 
@@ -636,11 +660,14 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
 
         for (i = 0; i < nprocs; i++) {
 #ifdef RDCOLL_DEBUG
-            DBG_FPRINTF(stderr, "rank %d, i %d, others_count %d\n", rank, i, others_req[i].count);
+            DBG_FPRINTF(stderr, "rank %d, i %d, others_count %lld\n", rank, i, others_req[i].count);
 #endif
             if (others_req[i].count) {
                 start_pos[i] = curr_offlen_ptr[i];
+                MPI_Count j = 0;
                 for (j = curr_offlen_ptr[i]; j < others_req[i].count; j++) {
+                    ADIO_Offset req_off;
+                    MPI_Aint req_len;
                     if (partial_send[i]) {
                         /* this request may have been partially
                          * satisfied in the previous iteration. */
@@ -658,17 +685,13 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
                         count[i]++;
                         ADIOI_Assert((((ADIO_Offset) (uintptr_t) read_buf) + req_off - real_off) ==
                                      (ADIO_Offset) (uintptr_t) (read_buf + req_off - real_off));
-                        MPI_Get_address(read_buf + req_off - real_off,
-                                        &(others_req[i].mem_ptrs[j]));
-                        ADIOI_Assert((real_off + real_size - req_off) ==
-                                     (int) (real_off + real_size - req_off));
-                        send_size[i] +=
-                            (int) (MPL_MIN
-                                   (real_off + real_size - req_off,
-                                    (ADIO_Offset) (unsigned) req_len));
+                        MPI_Aint addr;
+                        MPI_Get_address(read_buf + req_off - real_off, &addr);
+                        others_req[i].mem_ptrs[j] = addr;
+                        send_size[i] += (MPL_MIN(real_off + real_size - req_off, req_len));
 
-                        if (real_off + real_size - req_off < (ADIO_Offset) (unsigned) req_len) {
-                            partial_send[i] = (int) (real_off + real_size - req_off);
+                        if (real_off + real_size - req_off < req_len) {
+                            partial_send[i] = (real_off + real_size - req_off);
                             if ((j + 1 < others_req[i].count) &&
                                 (others_req[i].offsets[j + 1] < real_off + real_size)) {
                                 /* this is the case illustrated in the
@@ -688,29 +711,17 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
             }
         }
 
-        flag = 0;
-        for (i = 0; i < nprocs; i++)
-            if (count[i])
-                flag = 1;
-
-        if (flag) {
-            ADIOI_Assert(size == (int) size);
-            ADIOI_LUSTRE_RD_LOCK_AHEAD(fd, fd->hints->cb_nodes, off, error_code);
-            ADIO_ReadContig(fd, read_buf + for_curr_iter, (int) size, MPI_BYTE,
-                            ADIO_EXPLICIT_OFFSET, off, &status, error_code);
-            if (*error_code != MPI_SUCCESS)
-                return;
-        }
-
         for_curr_iter = for_next_iter;
 
+        MPI_Aint recved_bytes = 0;
         ADIOI_R_Exchange_data(fd, buf, flat_buf, offset_list, len_list,
                               send_size, recv_size, count,
                               start_pos, partial_send, recd_from_proc, nprocs,
                               myrank,
                               buftype_is_contig, contig_access_count,
                               min_st_offset, fd_size, fd_start, fd_end,
-                              others_req, m, buftype_extent, buf_idx);
+                              others_req, m, buftype_extent, buf_idx, &recved_bytes);
+        actual_recved_bytes += recved_bytes;
 
 
         if (for_next_iter) {
@@ -732,32 +743,38 @@ static void ADIOI_Read_and_exch(ADIO_File fd, void *buf, MPI_Datatype
 
     for (i = 0; i < nprocs; i++)
         count[i] = send_size[i] = 0;
-    for (m = ntimes; m < max_ntimes; m++)
-/* nothing to send, but check for recv. */
+    for (m = ntimes; m < max_ntimes; m++) {
+        /* nothing to send, but check for recv. */
+        MPI_Aint recved_bytes = 0;
         ADIOI_R_Exchange_data(fd, buf, flat_buf, offset_list, len_list,
                               send_size, recv_size, count,
                               start_pos, partial_send, recd_from_proc, nprocs,
                               myrank,
                               buftype_is_contig, contig_access_count,
                               min_st_offset, fd_size, fd_start, fd_end,
-                              others_req, m, buftype_extent, buf_idx);
+                              others_req, m, buftype_extent, buf_idx, &recved_bytes);
+        actual_recved_bytes += recved_bytes;
+    }
+
+#ifdef HAVE_STATUS_SET_BYTES
+    MPIR_Status_set_bytes(status, datatype, actual_recved_bytes);
+#endif
 
     ADIOI_Free(curr_offlen_ptr);
 }
 
 static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                                   * flat_buf, ADIO_Offset * offset_list, ADIO_Offset
-                                  * len_list, int *send_size, int *recv_size,
-                                  int *count, int *start_pos, int *partial_send,
-                                  int *recd_from_proc, int nprocs,
-                                  int myrank, int
-                                  buftype_is_contig, int contig_access_count,
+                                  * len_list, MPI_Count * send_size, MPI_Count * recv_size,
+                                  MPI_Count * count, MPI_Count * start_pos,
+                                  MPI_Count * partial_send, MPI_Count * recd_from_proc, int nprocs,
+                                  int myrank, int buftype_is_contig, MPI_Count contig_access_count,
                                   ADIO_Offset min_st_offset, ADIO_Offset fd_size,
                                   ADIO_Offset * fd_start, ADIO_Offset * fd_end,
-                                  ADIOI_Access * others_req,
-                                  int iter, MPI_Aint buftype_extent, MPI_Aint * buf_idx)
+                                  ADIOI_Access * others_req, int iter, MPI_Aint buftype_extent,
+                                  MPI_Aint * buf_idx, MPI_Aint * actual_recved_bytes)
 {
-    int i, j, k = 0, tmp = 0, nprocs_recv, nprocs_send;
+    int i, nprocs_recv, nprocs_send;
     char **recv_buf = NULL;
     size_t memLen;
     MPI_Request *requests;
@@ -767,7 +784,7 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
 /* exchange send_size info so that each process knows how much to
    receive from whom and how much memory to allocate. */
 
-    MPI_Alltoall(send_size, 1, MPI_INT, recv_size, 1, MPI_INT, fd->comm);
+    MPI_Alltoall(send_size, 1, MPI_COUNT, recv_size, 1, MPI_COUNT, fd->comm);
 
     nprocs_recv = 0;
     nprocs_send = 0;
@@ -791,12 +808,12 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
     MPE_Log_event(5032, 0, NULL);
 #endif
 
+    MPI_Count j = 0; // think of this as a counter of non-zero sends/recs
     if (buftype_is_contig) {
-        j = 0;
         for (i = 0; i < nprocs; i++) {
             if (recv_size[i]) {
-                MPI_Irecv(((char *) buf) + buf_idx[i], recv_size[i],
-                          MPI_BYTE, i, ADIOI_COLL_TAG(i, iter), fd->comm, requests + j);
+                MPI_Irecv_c(((char *) buf) + buf_idx[i], recv_size[i],
+                            MPI_BYTE, i, ADIOI_COLL_TAG(i, iter), fd->comm, requests + j);
                 j++;
                 buf_idx[i] += recv_size[i];
             }
@@ -811,11 +828,11 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
         j = 0;
         for (i = 0; i < nprocs; i++) {
             if (recv_size[i]) {
-                MPI_Irecv(recv_buf[i], recv_size[i], MPI_BYTE, i,
-                          ADIOI_COLL_TAG(i, iter), fd->comm, requests + j);
+                MPI_Irecv_c(recv_buf[i], recv_size[i], MPI_BYTE, i,
+                            ADIOI_COLL_TAG(i, iter), fd->comm, requests + j);
                 j++;
 #ifdef RDCOLL_DEBUG
-                DBG_FPRINTF(stderr, "node %d, recv_size %d, tag %d \n",
+                DBG_FPRINTF(stderr, "node %d, recv_size %lld, tag %d \n",
                             myrank, recv_size[i], ADIOI_COLL_TAG(i, iter));
 #endif
             }
@@ -824,10 +841,12 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
 
 /* create derived datatypes and send data */
 
-    j = 0;
+    j=0;
     for (i = 0; i < nprocs; i++) {
         if (send_size[i]) {
-/* take care if the last off-len pair is a partial send */
+            /* take care if the last off-len pair is a partial send */
+            ADIO_Offset tmp = 0;
+            MPI_Count k = 0;
             if (partial_send[i]) {
                 k = start_pos[i] + count[i] - 1;
                 tmp = others_req[i].lens[k];
@@ -848,27 +867,33 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
         }
     }
 
-#ifdef MPI_STATUSES_IGNORE
-    statuses = MPI_STATUSES_IGNORE;
-#else
-    statuses = (MPI_Status *) ADIOI_Malloc((nprocs_send + nprocs_recv + 1) * sizeof(MPI_Status));
     /* +1 to avoid a 0-size malloc */
-#endif
+    statuses = (MPI_Status *) ADIOI_Malloc((nprocs_send + nprocs_recv + 1) * sizeof(MPI_Status));
 
     /* wait on the receives */
     if (nprocs_recv) {
 #ifdef NEEDS_MPI_TEST
-        j = 0;
+	j=0;
         while (!j)
             MPI_Testall(nprocs_recv, requests, &j, statuses);
 #else
         MPI_Waitall(nprocs_recv, requests, statuses);
 #endif
+        *actual_recved_bytes = 0;
+	j=0;
+        for (i = 0; i < nprocs; i++) {
+            if (recv_size[i]) {
+                MPI_Count count_recved;
+                MPI_Get_count_c(&statuses[j], MPI_BYTE, &count_recved);
+                *actual_recved_bytes += count_recved;
+                j++;
+            }
+        }
 
         /* if noncontiguous, to the copies from the recv buffers */
         if (!buftype_is_contig)
             ADIOI_Fill_user_buffer(fd, buf, flat_buf, recv_buf,
-                                   offset_list, len_list, (unsigned *) recv_size,
+                                   offset_list, len_list, recv_size,
                                    requests, statuses, recd_from_proc,
                                    nprocs, contig_access_count,
                                    min_st_offset, fd_size, fd_start, fd_end, buftype_extent);
@@ -879,9 +904,9 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
     MPI_Waitall(nprocs_send, requests + nprocs_recv, MPI_STATUSES_IGNORE);
 #else
     MPI_Waitall(nprocs_send, requests + nprocs_recv, statuses + nprocs_recv);
+#endif
 
     ADIOI_Free(statuses);
-#endif
     ADIOI_Free(requests);
 
     if (!buftype_is_contig) {
@@ -944,10 +969,10 @@ static void ADIOI_R_Exchange_data(ADIO_File fd, void *buf, ADIOI_Flatlist_node
 void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                             * flat_buf, char **recv_buf, ADIO_Offset
                             * offset_list, ADIO_Offset * len_list,
-                            unsigned *recv_size,
+                            MPI_Count * recv_size,
                             MPI_Request * requests, MPI_Status * statuses,
-                            int *recd_from_proc, int nprocs,
-                            int contig_access_count,
+                            MPI_Count * recd_from_proc, int nprocs,
+                            MPI_Count contig_access_count,
                             ADIO_Offset min_st_offset,
                             ADIO_Offset fd_size, ADIO_Offset * fd_start,
                             ADIO_Offset * fd_end, MPI_Aint buftype_extent)
@@ -955,12 +980,11 @@ void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
 
 /* this function is only called if buftype is not contig */
 
-    int i, p, flat_buf_idx;
+    int p, flat_buf_idx;
     ADIO_Offset flat_buf_sz, size_in_buf, buf_incr, size;
     int n_buftypes;
     ADIO_Offset off, len, rem_len, user_buf_idx;
-    /* Not sure unsigned is necessary, but it makes the math safer */
-    unsigned *curr_from_proc, *done_from_proc, *recv_buf_idx;
+    MPI_Count *curr_from_proc, *done_from_proc, *recv_buf_idx;
 
     MPL_UNREFERENCED_ARG(requests);
     MPL_UNREFERENCED_ARG(statuses);
@@ -971,11 +995,13 @@ void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                         filled into user buffer in previous iterations
     user_buf_idx = current location in user buffer
     recv_buf_idx[p] = current location in recv_buf of proc. p  */
-    curr_from_proc = (unsigned *) ADIOI_Malloc(nprocs * 3 * sizeof(unsigned));
+    /* combining these three related arrays into a single memory allocation
+     * (the "times 3" here) can help some highly noncontiguous workloads a bit */
+    curr_from_proc = ADIOI_Malloc(nprocs * 3 * sizeof(*curr_from_proc));
     done_from_proc = curr_from_proc + nprocs;
     recv_buf_idx = done_from_proc + nprocs;
 
-    for (i = 0; i < nprocs; i++) {
+    for (int i = 0; i < nprocs; i++) {
         recv_buf_idx[i] = curr_from_proc[i] = 0;
         done_from_proc[i] = recd_from_proc[i];
     }
@@ -989,7 +1015,7 @@ void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
      * flat_buf_sz = size of current contiguous component in
      * flattened buf */
 
-    for (i = 0; i < contig_access_count; i++) {
+    for (MPI_Count i = 0; i < contig_access_count; i++) {
         off = offset_list[i];
         rem_len = len_list[i];
 
@@ -1009,20 +1035,14 @@ void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                                        done_from_proc[p], recv_size[p] - recv_buf_idx[p]);
                         buf_incr = done_from_proc[p] - curr_from_proc[p];
                         ADIOI_BUF_INCR buf_incr = curr_from_proc[p] + len - done_from_proc[p];
-                        ADIOI_Assert((done_from_proc[p] + size) ==
-                                     (unsigned) ((ADIO_Offset) done_from_proc[p] + size));
                         curr_from_proc[p] = done_from_proc[p] + size;
                     ADIOI_BUF_COPY} else {
                         size = MPL_MIN(len, recv_size[p] - recv_buf_idx[p]);
                         buf_incr = len;
-                        ADIOI_Assert((curr_from_proc[p] + size) ==
-                                     (unsigned) ((ADIO_Offset) curr_from_proc[p] + size));
-                        curr_from_proc[p] += (unsigned) size;
+                        curr_from_proc[p] += size;
                     ADIOI_BUF_COPY}
                 } else {
-                    ADIOI_Assert((curr_from_proc[p] + len) ==
-                                 (unsigned) ((ADIO_Offset) curr_from_proc[p] + len));
-                    curr_from_proc[p] += (unsigned) len;
+                    curr_from_proc[p] += len;
                     buf_incr = len;
                 ADIOI_BUF_INCR}
             } else {
@@ -1032,7 +1052,7 @@ void ADIOI_Fill_user_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
             rem_len -= len;
         }
     }
-    for (i = 0; i < nprocs; i++)
+    for (int i = 0; i < nprocs; i++)
         if (recv_size[i])
             recd_from_proc[i] = curr_from_proc[i];
 
